from typing import Dict

original_localization: Dict[str, str] = {
    'global.quitConfirmation': 'Do you really want to quit?',
    'global.ok': 'OK',
    'global.quit': 'Quit',
    'global.cancel': 'Cancel',
    'global.saveFile': 'Save file',
    'cocoa.menu.about': 'About',
    'cocoa.menu.services': 'Services',
    'cocoa.menu.view': 'View',
    'cocoa.menu.edit': 'Edit',
    'cocoa.menu.hide': 'Hide',
    'cocoa.menu.hideOthers': 'Hide Others',
    'cocoa.menu.showAll': 'Show All',
    'cocoa.menu.quit': 'Quit',
    'cocoa.menu.fullscreen': 'Enter Fullscreen',
    'cocoa.menu.cut': 'Cut',
    'cocoa.menu.copy': 'Copy',
    'cocoa.menu.paste': 'Paste',
    'cocoa.menu.selectAll': 'Select All',
    'windows.fileFilter.allFiles': 'All files',
    'windows.fileFilter.otherFiles': 'Other file types',
    'linux.openFile': 'Open file',
    'linux.openFiles': 'Open files',
    'linux.openFolder': 'Open folder',
}
